# JamminLobby

JamminLobby is a powerful multiplayer lobby system for Godot 4.x that provides:

- Seamless multiplayer game hosting and joining
- Automatic LAN game discovery
- Player state synchronization and persistence
- Built-in chat system
- Options management and persistence
- Request/response system for client-server communication

You provide your own lobby UI, and the plugin will handle the hosting, joining, and more.

## Installation

1. Copy the `jammin_lobby` folder into your project's `addons/` directory
2. Enable the plugin in Project Settings -> Plugins

## Core Concepts

### The Lobby Singleton

The `Lobby` singleton (which is automatically added as an autoload) provides the main interface for multiplayer functionality:

```gdscript
# Configure the lobby
Lobby.setup({
    "lobby_name": "My Game",
    "game_port": 1234,
    "broadcast_port": 1235,
    "response_port": 1236,
    "player_script": preload("res://Scripts/MyPlayer.gd"),
    "autosave": true,
    "save_slot": 1,
    "options_save_file": "user://options.json"
})

# Host a game
Lobby.start_hosting()

# Join a game at a specific IP and port
Lobby.join({
    "ip": "192.168.1.100",
    "port": 1234
})

# Leave the current game
Lobby.leave("Player left the game")

# Send chat messages
Lobby.send_chat("Hello everyone!")
Lobby.send_system_chat("System message")
```

### Players

Players are managed through a custom player class that extends `JamminPlayer`. The plugin automatically handles player spawning, despawning, and state synchronization.

```gdscript
class_name MyPlayer extends JamminPlayer

var color: Color = Color.WHITE
var character: String = "fighter"

func _init():
    # Properties to sync across the network
    sync_prop("color")
    sync_prop("character")

    # Properties to save locally
    save_props(["color", "character"])

func on_me_created():
    # Called when local player is created
    pass

func on_joined_lobby():
    # Called when player joins a lobby
    pass

func on_left_lobby():
    # Called when player leaves a lobby
    pass
```

### Options Management

JamminLobby provides a robust options management system that is persisted to a file between game sessions:

```gdscript
# Global lobby options
Lobby.options.set("music_volume", 0.5)
Lobby.options.get("music_volume")

# Per-player options
Lobby.me.options.set("controls_inverted", true)
Lobby.me.options.get("controls_inverted")
```

### Request/Response System

The built-in request system allows for reliable client-server communication:

```gdscript
# On the host
Lobby.on_ask("start_game", func(data, requester):
    # Handle start game request
    return {"result": "ok"}
)

# On the client
var response = await Lobby.ask(Lobby.SERVER_ID, "start_game", {})
if response.result == "ok":
    start_game()
```

### Chat System

The chat system supports both regular (from a player) and system messages to the local player:

```gdscript
# Send a regular chat message
Lobby.send_chat("Hello everyone!")

# Send a system message (local only)
Lobby.send_system_chat("Game starting in 3...")

# Listen for chat updates
Lobby.chat_messages_updated.connect(func():
    for message in Lobby.chat_messages:
        print(message.body)
)
```

## Signals

### Core Signals

```gdscript
# Lobby state
signal joining_lobby()
signal joined_lobby()
signal left_lobby(reason: String)

# Player events
signal player_joining_lobby(pid: int)
signal player_joined_lobby(player: JamminPlayer)
signal player_left_lobby(player: JamminPlayer)
signal player_updated(player: JamminPlayer)

# Host events
signal hosting_started()
signal hosting_failed(message: String)
signal hosting_stopped(message: String)

# Discovery
signal discovery_server_started()
signal discovery_server_failed(error: int)
signal discovery_server_stopped()

# Chat
signal chat_messages_updated()
```

## Advanced Features

### Save Slots

JamminLobby supports multiple save slots for player data:

```gdscript
Lobby.save_slot = 2  # Switch to save slot 2
Lobby.me.restore()   # Restore player data from slot 2
Lobby.me.save()      # Save player data to slot 2
```

### Proxy Support

> Note: Proxy support is currently in development and not yet functional. This will allow for relay servers and NAT traversal in future versions.

## License

MIT License - see LICENSE.md for details.

Created by Jamon Holmgren ([Jammin Games](https://jammin.games))
